"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Validator_1 = require("./Validator");
class Validation extends Validator_1.Validator {
    constructor(rules, allowNull = false, allowUndefined = false, invertNext = false) {
        super(rules, allowNull, allowUndefined);
        this.invertNext = invertNext;
    }
    withRule(rule) {
        if (this.invertNext) {
            const orig = rule;
            rule = { id: 'not', rule: orig, describe: () => `not ${orig.describe()}`, test: v => !orig.test(v) };
        }
        return super.withRule(rule);
    }
    array() {
        return this.withRule({ id: 'array', describe: () => 'array', test: (v) => Array.isArray(v) });
    }
    boolean() {
        return this.withRule({ id: 'boolean', describe: () => 'boolean', test: (v) => typeof v === 'boolean' });
    }
    integer() {
        return this.withRule({ id: 'integer', describe: () => 'integer', test: (v) => Number.isInteger(v) });
    }
    null() {
        return this.withRule({ id: 'null', describe: () => 'null', test: (v) => v === null });
    }
    number() {
        return this.withRule({ id: 'number', describe: () => 'number', test: (v) => typeof v === 'number' });
    }
    numeric() {
        return this.withRule({ id: 'numeric', describe: () => 'numeric', test: (v) => isNumeric(v) });
    }
    object() {
        return this.withRule({ id: 'object', describe: () => 'object', test: (v) => isObject(v) });
    }
    string() {
        return this.withRule({ id: 'string', describe: () => 'string', test: (v) => typeof v === 'string' });
    }
    undefined() {
        return this.withRule({ id: 'undefined', describe: () => 'undefined', test: (v) => v === undefined });
    }
    get not() {
        return new Validation(this.rules, this.allowNull, this.allowUndefined, true);
    }
    anyOf(...validators) {
        return this.withRule({
            id: 'anyOf',
            rulesAlternatives: validators.map(vd => vd.rules),
            describe: () => '(' + validators.map(vd => vd.describe()).join(' or ') + ')',
            test: (v) => validators.some(vd => vd.test(v))
        });
    }
    equal(value) {
        return this.withRule({
            id: 'equal',
            value,
            describe: () => `value equal to ${JSON.stringify(value)}`,
            test: (v) => v == value
        });
    }
    exact(value) {
        return this.withRule({
            id: 'exact',
            value,
            describe: () => `value exactly equal to ${JSON.stringify(value)}`,
            test: (v) => v === value
        });
    }
    predicate(test, description) {
        return this.withRule({
            id: 'predicate',
            describe: () => description,
            test
        });
    }
    between(min, max) {
        if (min > max) {
            throw new Error(`Invalid range constraint: minimum (${min}) is greater than maximum (${max})`);
        }
        return this.withRule({
            id: 'between',
            min,
            max,
            describe: () => `number between ${min} and ${max}`,
            test: (v) => isNumeric(v) && v >= min && v <= max
        });
    }
    greaterThan(bound) {
        return this.withRule({
            id: 'greaterThan',
            bound,
            describe: () => `number greater than ${bound}`,
            test: (v) => isNumeric(v) && v > bound
        });
    }
    greaterThanOrEqual(bound) {
        return this.withRule({
            id: 'greaterThanOrEqual',
            bound,
            describe: () => `number greater than or equal to ${bound}`,
            test: (v) => isNumeric(v) && v >= bound
        });
    }
    lessThan(bound) {
        return this.withRule({
            id: 'lessThan',
            bound,
            describe: () => `number less than ${bound}`,
            test: (v) => isNumeric(v) && v < bound
        });
    }
    lessThanOrEqual(bound) {
        return this.withRule({
            id: 'lessThanOrEqual',
            bound,
            describe: () => `number less than or equal to ${bound}`,
            test: (v) => isNumeric(v) && v <= bound
        });
    }
    positive() {
        return this.withRule({ id: 'positive', describe: () => 'positive number', test: (v) => v > 0 });
    }
    negative() {
        return this.withRule({ id: 'negative', describe: () => 'negative number', test: (v) => v < 0 });
    }
    length(min, max) {
        if (max == null) {
            return this.withRule({
                id: 'length',
                min,
                describe: () => `length of ${min}`,
                test: (v) => hasLength(v) && v.length === min
            });
        }
        if (min > max) {
            throw new Error(`Invalid length constraint: minimum (${min}) is greater than maximum (${max})`);
        }
        return this.withRule({
            id: 'length',
            min,
            max,
            describe: () => `length between ${min} and ${max}`,
            test: (v) => hasLength(v) && v.length >= min && v.length <= max
        });
    }
    minLength(min) {
        return this.withRule({
            id: 'minLength',
            min,
            describe: () => `minimum length of ${min}`,
            test: (v) => hasLength(v) && v.length >= min
        });
    }
    maxLength(max) {
        return this.withRule({
            id: 'maxLength',
            max,
            describe: () => `maximum length of ${max}`,
            test: (v) => hasLength(v) && v.length <= max
        });
    }
    empty() {
        return this.withRule({
            id: 'empty',
            describe: () => 'zero length',
            test: (v) => hasLength(v) && v.length === 0
        });
    }
    pattern(regex) {
        return this.withRule({
            id: 'pattern',
            regex,
            describe: () => `string matching ${regex}`,
            test: (v) => typeof v === 'string' && regex.test(v)
        });
    }
    contains(str) {
        return this.withRule({
            id: 'contains',
            str,
            describe: () => `string containing ${JSON.stringify(str)}`,
            test: (v) => typeof v === 'string' && v.includes(str)
        });
    }
    startsWith(str) {
        return this.withRule({
            id: 'startsWith',
            str,
            describe: () => `string starting with ${JSON.stringify(str)}`,
            test: (v) => typeof v === 'string' && v.startsWith(str)
        });
    }
    endsWith(str) {
        return this.withRule({
            id: 'endsWith',
            str,
            describe: () => `string ending with ${JSON.stringify(str)}`,
            test: (v) => typeof v === 'string' && v.endsWith(str)
        });
    }
    includes(value) {
        return this.withRule({
            id: 'includes',
            value,
            describe: () => `array containing ${JSON.stringify(value)}`,
            test: (v) => Array.isArray(v) && v.includes(value)
        });
    }
    every(elementValidator) {
        return this.withRule({
            id: 'every',
            rules: elementValidator.rules,
            describe: () => `array with every element ${elementValidator.describe()}`,
            test: (v) => Array.isArray(v) && v.every(e => elementValidator.test(e))
        });
    }
    some(elementValidator) {
        return this.withRule({
            id: 'some',
            rules: elementValidator.rules,
            describe: () => `array with some element ${elementValidator.describe()}`,
            test: (v) => Array.isArray(v) && v.some(e => elementValidator.test(e))
        });
    }
    instanceOf(ctor) {
        return this.withRule({
            id: 'instanceOf',
            ctor,
            describe: () => `instance of ${ctor.name}`,
            test: (v) => v instanceof ctor
        });
    }
    keys(arrayValidator) {
        return this.withRule({
            id: 'keys',
            rules: arrayValidator.rules,
            describe: () => `object with keys ${arrayValidator.describe()}`,
            test: (v) => isObject(v) && arrayValidator.test(Object.keys(v))
        });
    }
    values(arrayValidator) {
        return this.withRule({
            id: 'values',
            rules: arrayValidator.rules,
            describe: () => `object with values ${arrayValidator.describe()}`,
            test: (v) => isObject(v) && arrayValidator.test(Object.values(v))
        });
    }
}
exports.Validation = Validation;
function hasLength(v) {
    return typeof v === 'string' || Array.isArray(v);
}
function isNumeric(v) {
    return !isNaN(parseFloat(v));
}
function isObject(v) {
    return v && typeof v === 'object' && !Array.isArray(v);
}
//# sourceMappingURL=Validation.js.map